@@ -7,10 +7,13 @@ import pyodbc
 import sqlite3
 from typing import Optional, Iterator, Any, Union, Hashable
 
+Connection = Union[pyodbc.Connection, sqlite3.Connection]  # Todo: unify in own interface to get autocomplete
+Cursor = Union[pyodbc.Cursor, sqlite3.Cursor]
 Record = dict[str, Any]
+RecordList = list[Record]
 
 
-def get_db_connection(server: str, database: str) -> pyodbc.Connection:
+def get_db_connection(server: str, database: str) -> Connection:
     """
     Initiate pyodbc connection to a SQL Server database. This function is intended to be suitable for cloud development:
     in the cloud you use environment variables to log in under a certain username and password, whereas locally you
@@ -53,7 +56,7 @@ def get_db_connection(server: str, database: str) -> pyodbc.Connection:
 
 
 @contextmanager
-def cursor(db_server: str, db_name: str) -> Iterator[pyodbc.Cursor]:
+def cursor(db_server: str, db_name: str) -> Iterator[Cursor]:
     """
     Obtain a cursor for a certain database server and database name. Internally uses `get_db_connection`. Use this as
     a context manager, which will handle closing the cursor and the connection. NOTE: this will not handle transaction
@@ -72,35 +75,11 @@ def cursor(db_server: str, db_name: str) -> Iterator[pyodbc.Cursor]:
         c.close()
         conn.close()
 
-def with_cursor(db_server: str, db_name: str):
-    """
-    Decorator to fill the first argument to a function with a cursor. This is useful when you want a piece of code to
-    have access over a longer period of time, for instance in an Azure Funciton. NOTE: this will not handle transaction
-    support, most of the time that means you need to commit your transactions yourself!
-    Example usage:
-    ```
-    @with_cursor('my_server.net', 'test')
-    def write_stuff(c, stuff):
-        c.execute('insert into test_table (x, y) values (?, ?)', stuff)
-        c.connection.commit()
 
-    write_stuff((1, 2))
-    ```
-    """
-    def decorator(f):
-        def wrapped(*a, **kw):
-            with cursor(db_server, db_name) as c:
-                return f(c, *a, **kw)
-        return wrapped
-    return decorator
-
-
-def query(c: pyodbc.Cursor, sql: str, data: Optional[tuple] = None) -> list[dict[str, Any]]:
+def query(c: Cursor, sql: str, data: Optional[tuple] = None) -> RecordList:
     """
     Call `c.execute(sql, data).fetchall()` and format the resulting rowset a list of records of the form
     [{colname: value}].
-
-    NOTE: while the type of `c` is a pyodbc Cursor, any DB API 2.0 conforming cursor will work.
     """
     if data is None:
         result = c.execute(sql).fetchall()
@@ -110,18 +89,55 @@ def query(c: pyodbc.Cursor, sql: str, data: Optional[tuple] = None) -> list[dict
     return [dict(zip(headers, r)) for r in result]
 
 
-def get_all(c: pyodbc.Cursor, table_name: str) -> list[dict[str, Any]]:
+def get_all(c: Cursor, table_name: str) -> RecordList:
     """
     Get all current data from table `table_name`.
 
     IMPORTANT WARNING: `table_name` is not sanitized. Don't pass untrusted table names to this function!
-    NOTE: while the type of `c` is a pyodbc Cursor, any DB API 2.0 conforming cursor will work.
     """
     return query(c, f'select * from {table_name}')
 
 
-def write(c: pyodbc.Cursor, table_name: str, data: list[dict[str, Any]],
-          primary_key: Optional[Union[str, tuple]] = None, *, update=True, insert=True, delete=True):
+def validate(data: RecordList):
+    assert len(unique := set(tuple(sorted(r.keys())) for r in data)) == 1, \
+        f'Non-uniform list of dictionaries passed, got differing keys {unique}.'
+    assert not any(non_str := {k: type(k) for k in data[0].keys() if not isinstance(k, str)}), \
+        f'Non-string keys in data, got keys with types {non_str}.'
+
+
+def insert_with_return(
+        c: Cursor, table_name: str, data: Record, return_columns: Optional[Union[str, tuple]] = None
+) -> Record:
+    """
+    Insert data into the database, returning a set of return columns. The primary use for this is if you have columns
+    generated by your database, like an identity. Returns input record with returned columns added (if any).
+    """
+    validate([data])
+    return_columns = (return_columns,) if isinstance(return_columns, str) else return_columns
+
+    columns = data.keys()
+    insert_data = tuple(data[col] for col in columns)
+    text_columns = ', '.join(columns)
+    placeholders = ', '.join('?' for _ in columns)
+    # Dispatch on cursor type for now, pyodbc type is for MSSQL only
+    if return_columns is None:
+        sql = f'insert into {table_name}({text_columns}) values ({placeholders})'
+        c.execute(sql)
+        return data
+    elif isinstance(c, sqlite3.Cursor):
+        text_return_columns = ', '.join(return_columns)
+        sql = f'insert into {table_name}({text_columns}) values ({placeholders}) returning {text_return_columns}'
+        output = query(c, sql, insert_data)[0]
+        return {**data, **output}
+    else:
+        text_return_columns = ', '.join(f'Inserted.{col}' for col in return_columns)
+        sql = f'insert into {table_name}({text_columns}) output {text_return_columns} values ({placeholders})'
+        output = query(c, sql, insert_data)[0]
+        return {**data, **output}
+
+
+def write(c: Cursor, table_name: str, data: RecordList, primary_key: Optional[Union[str, tuple]] = None, *,
+          update=True, insert=True, delete=True):
     """
     Update data in database table. We check identity based on the keys of the IndexedPyFrame.
     `update`, `insert`, and `delete` control which actions to take. By default, this function emits the correct update,
@@ -130,21 +146,18 @@ def write(c: pyodbc.Cursor, table_name: str, data: list[dict[str, Any]],
     - `insert=True` means rows not already in the database will be added from the in-memory data
     - `delete=True` means rows present in the database but not in the in-memory database will be deleted
 
-    If `primary_key` is None, we will only execute an insert with the in-memory data.
+    If primary_key is None, only inserting is supported.
 
     IMPORTANT WARNING: `table_name` is not sanitized. Don't pass untrusted table names to this function!
-    NOTE: while the type of `c` is a pyodbc Cursor, any DB API 2.0 conforming cursor will work.
     """
-    # Data checks
-    assert len(unique := set(tuple(sorted(r.keys())) for r in data)) == 1, \
-        f'Non-uniform list of dictionaries passed, got differing keys {unique}.'
+    validate(data)
 
-    # Fix up primary key, set index, get a list of writeable columns, and read out in-database keys
+    # Deal with primary key, list of writeable columns, indexed data, data in db
     if primary_key is None:
-        assert not update and not delete, f'Need a primary key to update or delete'
+        assert not update and not delete, 'updating and deleting without specifying a primary key not supported'
         primary_key = tuple()
-        columns = tuple(data[0].keys())
         data = {i: r for i, r in enumerate(data)}
+        columns = tuple(k for k in data[0].keys())
         in_db = set()
     else:
         primary_key = (primary_key,) if isinstance(primary_key, str) else tuple(primary_key)
@@ -153,9 +166,13 @@ def write(c: pyodbc.Cursor, table_name: str, data: list[dict[str, Any]],
             logging.warning(f'Columns {empty_strings} contain empty strings. '
                             f'Generally inserting empty strings into a database is a bad idea.')
 
+        # List of writeable columns (for updates we don't try to overwrite the primary key)
         columns = tuple(k for k in data[0].keys() if k not in primary_key)
+
+        # Indexed data on primary key
         data = {tuple(r[i] for i in primary_key): r for r in data}
 
+        # Data present in database
         sql = f'select {", ".join(primary_key)} from {table_name}'
         in_db = {tuple(r[k] for k in primary_key) for r in query(c, sql)}
 
@@ -174,8 +191,7 @@ def write(c: pyodbc.Cursor, table_name: str, data: list[dict[str, Any]],
     if insert and (insert_keys := data.keys() - in_db):
         insert_data = [tuple(data[k][col] for col in columns + primary_key) for k in insert_keys]
 
-        # Use keyword placeholders, so we can pass a record to executemany
-        placeholders = ', '.join(f'?' for name in columns + primary_key)
+        placeholders = ', '.join(f'?' for _ in columns + primary_key)
         text_columns = ', '.join(columns + primary_key)
         sql = f'insert into {table_name}({text_columns}) VALUES ({placeholders})'
 
